/*
mylib helper library for Casio 9860 sdk v0.45
Copyright (C) 2011 Toussis Manolis (koppermind at gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#include "mylib.h"


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include "fxlib.h"


#ifdef MYLIB_DEBUG
char debugB[128]; //debugBuf
unsigned int debugkey; 
unsigned char* debugScr;
#endif //MYLIB_DEBUG

#define SCA 0xD201D002
#define SCB 0x422B0009
#define SCE 0x80010070
typedef char*(*sc_charp_short_shortp)(short,short*);
//typedef int(*sc_int_int)(int);
typedef int(*sc_charp_charp_ushortp_ushortp)(unsigned char*,unsigned char*,unsigned short*,unsigned short*);
typedef int(*sc_int_void)(void);

// char* GetMiniGlyph(short glyph,short* width)
const unsigned int sc0137[] = { SCA, SCB, SCE, 0x0137 };
#define GetMiniGlyph (*(sc_charp_short_shortp)sc0137)
/*
int Heap_SetTopChunk(int size);
const unsigned int sc0499[] = { SCA ,SCB,SCE, 0x0499 };
#define Heap_SetTopChunk (*(sc_int_int)sc0499)
*/
//int GlibGetOSVersionInfo(unsigned char *a, unsigned char *b, unsigned short *c, unsigned short *d);
const unsigned int sc0015[] = {SCA ,SCB,SCE,0x0015};
#define GlibGetOSVersionInfo (*(sc_charp_charp_ushortp_ushortp)sc0015)

const unsigned int sc003b[] = {SCA,SCB,SCE,0x003b};
#define RTC_GetTicks (*(sc_int_void)sc003b)

#define min(a,b) (a<=b ? (a) : (b))
#define max(a,b) (a>=b ? (a) : (b))

int GetRealTimeTicks(void )
{
  return RTC_GetTicks();
}


int GetOSVersion(void )
{
  unsigned char mainversion;
  unsigned char minorversion;
  unsigned short release;
  unsigned short build;
  GlibGetOSVersionInfo(&mainversion, &minorversion, &release, &build); //GlibGetOSVersionInfo = syscall 0x0015
  return ((mainversion << 24) & 0xFF000000) | ((minorversion << 16) & 0x00FF0000) | (release & 0x0000FFFF);
}


int GetMiniTextWidth(const char* text)
{
  int i=0;
  int px=0;
  unsigned char c;
  unsigned short glyph;
  short w;
  c=*(char*)(text+i);
  while(c!=0)
  {
    c=*(char*)(text+i);
    if(c==0x7F || c==0xE5 || c==0xE6)
    {
      i++;
      glyph=((unsigned short)(c)*256)+(unsigned char)(*(char*)(text+i));
      GetMiniGlyph(glyph,&w);
      px=px+w;
      i++;
    }else
    {
      glyph=(unsigned short)c;
      GetMiniGlyph(glyph,&w);
      px=px+w;
      i++;
    };
  };
  return px;
}
/*
typedef struct {
  unsigned int FreePieceLength;
  void* next;
} THeapFreePieceLinkedListItem;
int GetFreeMemory(void)
{
  int result,freespace;
  unsigned int linkliststart=0x88024000;
  THeapFreePieceLinkedListItem* FreePiece;
  result=Heap_SetTopChunk(0);
  result=0x88030000 - result;
  freespace=result;
  FreePiece=(THeapFreePieceLinkedListItem*)(linkliststart);
  while(FreePiece->next)
  {
    freespace = freespace + FreePiece->FreePieceLength;
    FreePiece=(THeapFreePieceLinkedListItem*)(FreePiece->next);
  };
  freespace = freespace + FreePiece->FreePieceLength;
  return freespace;
}
*/

unsigned char* GetScreen(void)
{
  unsigned char* p;
  p=(unsigned char *)malloc(1024); //1024 bytes size
  ASSERT(p!=0);
  Bdisp_PutDisp_DD();
  Bdisp_GetDisp_DD(p);
  return(p);
}


void SetScreen(unsigned char* p)
{
  GRAPHDATA g;
  DISPGRAPH d;
  g.height=64;g.width=128;g.pBitmap=p;
  d.x=0;d.y=0;d.GraphData=g;d.WriteModify=IMB_WRITEMODIFY_NORMAL;d.WriteKind=IMB_WRITEKIND_OVER;
  Bdisp_WriteGraph_DDVRAM(&d);
}

/* DOC: String struct */

struct String* stringInit(const char* text,int maxsize)
{
  int n;
  struct String* str;
  char* bufp;
  str=malloc(sizeof(struct String));
  ASSERT(str!=0);
  n=strlen(text);
  if(maxsize>n) n=maxsize; 
  str->data=(char*)malloc(sizeof(char)*(n+1));
  ASSERT(str->data!=0);
  str->maxLength=n;
  strcpy(str->data,text);
  return str;
}

char* stringGet(struct String* str)
{
  return str->data;
}

double stringGetDouble(struct String* str)
{
  double d;
  d=atof(str->data);
  return d;
}

int stringGetInteger(struct String* str)
{
  int i;
  i=atoi(str->data);
  return i;
}

void stringSet(struct String* str, const char* text)
{
  int n;
  n=strlen(text);
  if(n>(str->maxLength))//needs reallocation
  {
    free(str->data);
    str->data=(char*)malloc(sizeof(char)*(n+1));
    str->maxLength=n;
  };
  strcpy(str->data,text);
}

void stringAppend(struct String* str, const char* text)
{
  int n1,n2;
  char* bufp;
  n1=stringLength(str);n2=strlen(text);
  if(n1+n2>(str->maxLength))//needs reallocation
  {
    bufp=(char*)malloc(sizeof(char)*(n1+n2+1));
    str->maxLength=n1+n2;
    strcpy(bufp,str->data);
    free(str->data);
    strcat(bufp,text);
    str->data=bufp;
  }else
  {
    strcat(str->data,text);
  };
}

void stringFill(struct String* str, const char c, int n)
{
 if(n>str->maxLength)//needs realloc
 {
   free(str->data);
   str->data=(char*)malloc(sizeof(char)*(n+1));
 };
 memset(str->data,c,n);
 str->data[n]='\0';
}

void stringRemoveAt(struct String* str, int i, int n)
{
  struct String* s1;
  struct String* s2;
  int len=stringLength(str);
  s1=stringInit("",len);
  s2=stringInit("",len);
  stringMid(str,1,i-1,s1);//before
  stringMid(str,i+n,len,s2);//after
  stringAppend(s1,s2->data);
  stringSet(str,s1->data);
  stringFree(s1);
  stringFree(s2);
}


void stringCopy(struct String* dest,const struct String* src)
{
  stringSet(dest,src->data); 
}

void stringSetDouble(struct String* str, const double value,const char* format)
{
  if(format==0)
    stringSprintf(str,"%f",value); else stringSprintf(str,format,value);
}

void stringSetInteger(struct String* str,const int value,const char* format)
{
  if(format==0)
    stringSprintf(str,"%d",value); else stringSprintf(str,format,value);
}

int stringLength(struct String* str)
{
  return strlen(str->data);
}

void stringFree(struct String* str)
{
  free(str->data);
  free(str);
}

void stringSprintf(struct String* str, const char* format, ... )
{
  va_list ap;
  char buf[STRING_SPRINTF_SIZE];
  int n;
  va_start(ap,format);
  n=vsprintf(buf,format,ap);
  ASSERT(n<=(STRING_SPRINTF_SIZE-2) && n>=0);
  va_end(ap);
  stringSet(str,buf);
}

void stringFormatValue(struct String* str, const char* format, const char* value)
{
  double d1;
  int w1;
  unsigned int w2;
  char c;
  int len;
  len=strlen(format);
  if(memchr(format,'s',len)!=0)
  {
    stringSprintf(str,format,value);
  }else if(memchr(format,'d',len) || memchr(format,'i',len))//int
  {
    w1=atoi(value);
    stringSprintf(str,format,w1);
  }else if(memchr(format,'u',len)|| memchr(format,'o',len) 
    || memchr(format,'x',len)|| memchr(format,'X',len))
  {
    w2=atoi(value);
    stringSprintf(str,format,w2);
  }else if(memchr(format,'f',len) || memchr(format,'F',len)
    || memchr(format,'e',len) || memchr(format,'E',len)
    || memchr(format,'g',len) || memchr(format,'G',len))
  {
    d1=atof(value);
    stringSprintf(str,format,d1);
  }else if(memchr(format,'c',len))
  {
    c=value[0];
    stringSprintf(str,format,c);
  };
}


void stringMid(struct String* str, int i, int n, struct String* dest)
//i is 1based
{
  int len;
  ASSERT(i>=1);
  if(i>stringLength(str)) {stringSet(dest,"");return;};//returns ""
  if(i+n>stringLength(str)) len=stringLength(str); else len=n;
  if(dest->maxLength < len)//realloc
  {
    free(dest->data);
    dest->data=(char*)malloc(sizeof(char)*(len+1));
  };
  strncpy(dest->data,str->data+i-1,len);
  dest->data[len]=0;
  return;
}

int stringIdx(struct String* str, const char c, int instance)
{
  int len=stringLength(str);
  int i;
  int o=0;
  ASSERT(instance>=1);
  for(i=1;i<=len;i++)
  {
    if(str->data[i-1]==c) o++;
    if(o==instance) return i;
  }
  return -1;
}


void stringListItem(struct String* str, const char d, int i, struct String* dest)
{
  int a,b;
  ASSERT(i>=1);
  if(i==1)
  {
    a=1;
    b=stringIdx(str,d,i);
    if(b==-1) b=stringLength(str)+1;
  }else
  {
    a=stringIdx(str,d,i-1);
    b=stringIdx(str,d,i);
    if(a==-1) {a=1;b=1;goto end1;};
    if(b==-1) b=stringLength(str)+1;
    a++;
  };
  end1:
  stringMid(str,a,b-a,dest);
  return;
}

int stringListCount(struct String* str, const char d)
{
  int len=stringLength(str);
  int w1,i;
  i=0;
  for(w1=0;w1<len;w1++)
  {
    if(str->data[w1]==d) i++;
  };
  return(i+1);
}




void stringListItemModify(struct String* str, const char d, int i, const char* newitem, struct String* dest)
{
  struct String* item;
  char buf[2];
  int w1,n,no,w2;
  item=stringInit("",32);
  n=stringListCount(str,d);
  stringSet(dest,"");
  for(w1=1;w1<=n;w1++)
  {
    if(w1==i)
    {
      stringAppend(dest,newitem);
    }else
    {
      stringListItem(str,d,w1,item);
      stringAppend(dest,item->data);
    };
    buf[0]=d;buf[1]=0;
    if(w1!=n) stringAppend(dest,buf);
  };
  stringFree(item);
}


/* DialogItem */

struct DialogItem* dialogItemInit()
{
  struct DialogItem* itm;
  itm=malloc(sizeof(struct DialogItem));
  ASSERT(itm!=0);
  itm->type=0;
  itm->visible=1;
  itm->x=0;itm->y=0;itm->w=0;itm->h=0;itm->status=0;
  itm->current=0;
  itm->text=stringInit("",16);
  itm->format=stringInit("",8);
  itm->items=slistInit(0,8);
  itm->formats=slistInit(0,8);
  itm->values=slistInit(0,8);
  itm->next=0;
  return itm;
}

void dialogItemFree(struct DialogItem* item)
{
  stringFree(item->text);stringFree(item->format);
  slistFree(item->items);
  slistFree(item->formats);
  slistFree(item->values);
  free(item);
}

void dialogItemCopy(struct DialogItem* dest,const struct DialogItem* src, int nextlink)
{
   dest->type=src->type;
   dest->visible=src->visible;
   dest->x=src->x;dest->y=src->y;dest->w=src->w;dest->h=src->h;dest->status=src->status;
   dest->current=src->current;
   stringSet(dest->text,src->text->data);
   stringSet(dest->format,src->format->data);
   slistCopy(dest->items,src->items);
   slistCopy(dest->formats,src->formats);
   slistCopy(dest->values,src->values);
   if(nextlink==1) dest->next=src->next;
}

void dialogItemSetLabel(struct DialogItem* itm,int x,int y,const char* text)
{
  itm->type=DIALOG_LABEL;
  itm->visible=1;
  stringSet(itm->format,"");
  slistRemoveAll(itm->items);
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  itm->x=x;itm->y=y;
  stringSet(itm->text,text);
}

void dialogItemSetButton(struct DialogItem* itm,int x,int y,int w,int h,const char* text)
{
  itm->type=DIALOG_BUTTON;
  itm->visible=1;
  stringSet(itm->format,"");
  slistRemoveAll(itm->items);
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  itm->x=x;itm->y=y;itm->w=w;itm->h=h;
  stringSet(itm->text,text);
}

void dialogItemSetEditBox(struct DialogItem* itm,int x,int y,int w,int h,const char* text,const char* format)
{
  itm->type=DIALOG_EDITBOX;
  itm->visible=1;
  slistRemoveAll(itm->items);
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  stringSet(itm->text,text);
  stringSet(itm->format,format);
  if(h<9) h=9;//minimum height 
  itm->x=x;itm->y=y;itm->w=w;itm->h=h;
}

void dialogItemSetCheckBox(struct DialogItem* itm,int x,int y,const char* text,int status)
{
  itm->type=DIALOG_CHECKBOX;
  itm->visible=1;
  stringSet(itm->format,"");
  slistRemoveAll(itm->items);
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  stringSet(itm->text,text);
  itm->x=x;itm->y=y;itm->status=status;
}

void dialogItemSetRadio(struct DialogItem* itm,int x,int y,int current,const char* items)
{
  itm->type=DIALOG_RADIO;
  itm->visible=1;
  stringSet(itm->text,"");
  stringSet(itm->format,"");
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  slistSetAll(itm->items,',',items);
  itm->x=x;itm->y=y;itm->current=current;
}

void dialogItemSetDropList(struct DialogItem* itm,int x,int y,int w,int h,int current,const char* items)
{
  itm->type=DIALOG_DROPLIST;
  itm->visible=1;
  stringSet(itm->text,"");
  stringSet(itm->format,"");
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  slistSetAll(itm->items,',',items);
  if(h<9) h=9;//minimum height 
  itm->x=x;itm->y=y;itm->w=w;itm->h=h;itm->current=current;
}

void dialogItemSetList(struct DialogItem* itm, int x, int y, int w, int h, int current, const char* items)
{
  itm->type=DIALOG_LIST;
  itm->visible=1;
  stringSet(itm->text,"");
  stringSet(itm->format,"");
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  slistSetAll(itm->items,',',items);
  if(h<9) h=9;//minimum height 
  itm->x=x;itm->y=y;itm->w=w;itm->h=h;itm->current=current;
}


void dialogItemSetListEdit(struct DialogItem* itm,int x,int y,int w,int h,int current,const char* labels,const char* values,const char* formats,int readOnly)
{
  itm->type=DIALOG_LISTEDIT;
  itm->visible=1;
  stringSet(itm->text,"");
  stringSet(itm->format,"");
  slistSetAll(itm->items,',',labels);
  slistSetAll(itm->formats,',',formats);
  slistSetAll(itm->values,',',values);
  itm->status=readOnly;//if 1 cannot edit
  if(h<9) h=9;//minimum height 
  itm->x=x;itm->y=y;itm->w=w;itm->h=h;itm->current=current;
}

void dialogItemSetMenu(struct DialogItem* itm,const char* menu)
{
  itm->type=DIALOG_MENU;
  itm->visible=1;
  slistRemoveAll(itm->items);
  slistRemoveAll(itm->formats);
  slistRemoveAll(itm->values);
  stringSet(itm->text,menu);
  stringSet(itm->format,"");
}


void dialogPrepend(struct Dialog* dlg, const struct DialogItem* item)
{
  struct DialogItem* itm;
  struct DialogItem* tmp;
  tmp=dlg->first;
  itm=dialogItemInit();
  dialogItemCopy(itm,item,0);
  dlg->first=itm;
  itm->next=tmp; 
  dlg->count++;
}

struct Dialog* dialogInit()
{
  struct Dialog* dlg;
  dlg=malloc(sizeof(struct Dialog));
  ASSERT(dlg!=0);
  dlg->first=0;
  dlg->count=0;
  dlg->current=0;
  dlg->retCode=0;
  dlg->retSubcode=0;
  dlg->event=0;
  dlg->eventType=0;
  dlg->eventData=0;
  dlg->title=stringInit("",32);
  return dlg;
}

struct DialogItem* dialogGet(struct Dialog* dlg,int i)
{
  struct DialogItem* item;
  int j;
  j=1;
  item=dlg->first;
  ASSERT(item!=0);
  do {
    if(j==i) return(item);
    item=item->next;
    j++;
  } while(item!=0);
  ASSERT(-1);//to provoke an assert if there is no item
  return 0;
}



void dialogSet(struct Dialog* dlg, int i, struct DialogItem* item)
{
  struct DialogItem* itm;
  struct DialogItem* next;
  int j,len;
  j=1;
  itm=dlg->first;
  ASSERT(itm!=0);
  do {
    if(j==i)
    {
      dialogItemCopy(itm,item,0);
      return;
    };
    itm=itm->next;
    j++;
  } while(itm!=0);
  ASSERT(-1);// to provoke an assert if item not exist
}

void dialogFree(struct Dialog* dlg)
{
  struct DialogItem* item;
  struct DialogItem* tmp;
  item=dlg->first;
  while(item!=0)
  {
    tmp=item;
    item=item->next;
    dialogItemFree(tmp);
  };
  stringFree(dlg->title);
  free(dlg);
}

void dialogInsertAt(struct Dialog* dlg, int i, const struct DialogItem* item)
{
  struct DialogItem* itm;
  struct DialogItem* prev;
  struct DialogItem* newi;
  int j;
  itm=dlg->first;
  prev=0;
  j=1;
  do {
    if(i==j)//insert here
    {
      if(prev==0)
      {
	dialogPrepend(dlg,item);
      }else
      {
	newi=dialogItemInit();
	dialogItemCopy(newi,item,0);
	prev->next=newi;
	newi->next=itm;
	dlg->count++;
      };
      return;
    };
    prev=itm;
    itm=itm->next;
    j++;
  } while(itm!=0);
  ASSERT(-1); //to provoke assert if item position is not found
  return;
}

void dialogAppend(struct Dialog* dlg, const struct DialogItem* item)
{
  struct DialogItem* itm;
  struct DialogItem* newi;
  itm=dlg->first;
  if(itm!=0)
  {
   while(itm->next!=0)
   {
     itm=itm->next;
   };
   newi=dialogItemInit();
   dialogItemCopy(newi,item,0);
   itm->next=newi;
   dlg->count++;
  }else //list is empty
  {
    newi=dialogItemInit();
    dialogItemCopy(newi,item,0);
    dlg->first=newi;
    dlg->count++; 
  };
}

void dialogRemoveAll(struct Dialog* dlg)
{
  struct DialogItem* item;
  struct DialogItem* tmp;
  item=dlg->first;
  while(item!=0)
  {
    dlg->first=item->next;
    tmp=item;
    item=item->next;
    dialogItemFree(tmp);
    dlg->count--;
  };
}


void dialogRemoveAt(struct Dialog* dlg,int i)
{
  struct DialogItem* item;
  struct DialogItem* prev;
  int j;
  item=dlg->first;
  prev=0;
  j=1;
  do {
    if(i==j)//remove here
    {
      if(prev==0)
      {
	dlg->first=item->next;
	dialogItemFree(item);
      }else
      {
	prev->next=item->next;
	dialogItemFree(item);
      };
      dlg->count--;
      return;
    };
    prev=item;
    item=item->next;
    j++;
  } while(item!=0);
  return;
}



void dialogExec(struct Dialog* dlg, const char* title, int clearScreen)
{
	unsigned int key;
	int n,type;
	int w1,w2,w3,w4,w5,w6,x,y,w,h;
	struct String* s;
	struct String* s1;
	char* bufp1;
	char* bufp2;
	char* bufp3;
	int hasMenu;
	int MenuOption[6];
	struct DialogItem* menu=0;
	unsigned char* Scr1;
	unsigned char* Scr2;
	struct DialogItem* item;
	struct DialogItem* item2;
	struct StringList* sl;
	struct StringListNode* slit;
	
	s=stringInit("",32);
	s1=stringInit("",32);
	hasMenu=0;
	for(w1=0;w1<=5;w1++) MenuOption[w1]=0;
	dlg->current=1;
	n=dlg->count;
	Scr1=GetScreen();
	while(1)
	{
		if(clearScreen==0) SetScreen(Scr1); else Bdisp_AllClr_VRAM();
		stringSet(dlg->title,title);
		EVENT(dlg,EVENT_BEFORE_REFRESH,0);
		//draw Title
		if(strlen(title)>0) 
		{
		  stringSprintf(s,"< %s >",dlg->title->data);
		  PrintMini(64-GetMiniTextWidth(s->data)/2-1,0,(unsigned char*)s->data,MINI_OVER);
		};
		//draw each item
		for(w1=1;w1<=n;w1++)
		{
			item=dialogGet(dlg,w1);
			type=item->type;
			if(item->type==DIALOG_LABEL)// label
			{
			  	w2=0;
				if(w1==dlg->current) 
				{
				 jump3:
				  dlg->current++;w2++;//label does not keep focus
				  if(dlg->current>n) dlg->current=1;
				  item2=dialogGet(dlg,dlg->current);
				  if((item2->type==DIALOG_LABEL || item2->type==DIALOG_MENU || item2->visible==0) && w2<dlg->count) goto jump3;
				};
				if(item->visible==1)
				{
				  x=item->x;y=item->y;
				  bufp1=stringGet(item->text);
				  PrintMini(x,y,(unsigned char*)bufp1,MINI_OVER);
				};
			} else if(item->type==DIALOG_BUTTON && item->visible==1)//button
			{
				x=item->x;y=item->y;w=item->w;h=item->h;
				bufp1=stringGet(item->text);
				PrintMini(x+w/2-GetMiniTextWidth(bufp1)/2,y+h/2-3,(unsigned char*)bufp1,MINI_OVER);
				drawRect(x,y,x+w-2,y+h-2,1,STYLE_SOLID);
				drawRect(x,y,x+w-1,y+h-1,1,STYLE_SOLID);
				if(dlg->current==w1) Bdisp_AreaReverseVRAM(x+1,y+1,x+w-2,y+h-2);
			} else if(item->type==DIALOG_EDITBOX && item->visible==1)//editbox
			{
				x=item->x;y=item->y;w=item->w;h=item->h;
				bufp1=stringGet(item->text);
				PrintMini(x+w/2-GetMiniTextWidth(bufp1)/2-2,y+h/2-4,(unsigned char*)bufp1,MINI_OVER);
				drawRect(x-2,y-2,x+w-3,y+h-3,1,STYLE_SOLID);
				if(dlg->current==w1) Bdisp_AreaReverseVRAM(x,y,x+w-5,y+h-5);
			} else if(item->type==DIALOG_CHECKBOX && item->visible==1)//checkbox
			{
				x=item->x;y=item->y;
				bufp1=stringGet(item->text);
				PrintMini(x+9,y+2,(unsigned char*)bufp1,MINI_OVER);
				drawRect(x+1,y+1,x+7,y+7,1,STYLE_SOLID);
				if(item->status==1) //is checked
				{
					drawLine(x+2,y+2,x+6,y+6,1,STYLE_SOLID); // draw cross
					drawLine(x+2,y+6,x+6,y+2,1,STYLE_SOLID);
				};
				if(dlg->current==w1) Bdisp_AreaReverseVRAM(x,y,x+9+GetMiniTextWidth(bufp1)-1,y+8);
			} else if(item->type==DIALOG_RADIO && item->visible==1)//radio buttons 
			{
				x=item->x;y=item->y;
				w2=1;//max pixel width
				slit=item->items->first;w3=1;
				while(slit!=0)
				{
				  bufp1=slit->data;
				  w2=max(w2,GetMiniTextWidth(bufp1));
				  PrintMini(x+9,y+2+(w3-1)*7,(unsigned char*)bufp1,MINI_OVER);
				  drawCircle(x+4,y+4+(w3-1)*7,3,1);
				  w3++;
				  slit=slit->next;
				};
				drawCircle(x+4,y+4+(item->current-1)*7,1,1);
				Bdisp_SetPoint_VRAM(x+4,y+4+(item->current-1)*7,1);
				if(dlg->current==w1) Bdisp_AreaReverseVRAM(x,y,x+9+w2-1,y+7*item->items->count+1);
			}else if(item->type==DIALOG_DROPLIST && item->visible==1)//drop down list
			{
				x=item->x;y=item->y;w=item->w;h=item->h;
				sl=item->items;
				bufp1=slistGet(sl,item->current);
				x=x;y=y;
				PrintMini(x+w/2-GetMiniTextWidth(bufp1)/2-3,y+h/2-4,(unsigned char*)bufp1,MINI_OVER);
				drawRect(x-2,y-2,x+w-3,y+h-3,1,STYLE_SOLID);
				PrintMini(x+w-9,y,(unsigned char*)"\xE6\x93",MINI_OVER);
				if(dlg->current==w1) Bdisp_AreaReverseVRAM(x,y,x+w-5,y+h-5);
			}else if(item->type==DIALOG_LIST && item->visible==1)
			{
				x=item->x;y=item->y;w=item->w;h=item->h;
				sl=item->items;
				x=x-2;y=y-2;
				drawRect(x,y,x+w,y+h,1,STYLE_SOLID);
				w2=h/8;//number of lines
				w3=item->current-w2/2;if(w3<1) w3=1;//first item to draw
				w4=w3+w2-1;// last item to draw
				if(w4>sl->count) w4=sl->count;
				for(w5=w3;w5<=w4;w5++)
				{
				    bufp1=slistGet(sl,w5);
				    PrintMini(x+w/2-GetMiniTextWidth(bufp1)/2-1,y+2+(w5-w3)*8,(unsigned char*)bufp1,MINI_OVER);
				};
				drawRect(x+2,y+1+(item->current-w3)*8,x+w-8,y+(item->current-w3+1)*8-1,1,STYLE_SOLID);
				if(w4<sl->count) PrintMini(x+w-6,y+h-6,(unsigned char*)"\xE6\x93",MINI_OVER);
				if(w5>1) PrintMini(x+w-6,y+2,(unsigned char*)"\xE6\x92",MINI_OVER);
				//LIST is the only item that gets a rectangle instead of reverse
				if(dlg->current==w1) drawRect(x+1,y+1,x+w-1,y+h-1,1,STYLE_SOLID);
			}else if(item->type==DIALOG_LISTEDIT && item->visible==1)//LISTEDIT
			{
				x=item->x;y=item->y;w=item->w;h=item->h;
				sl=item->items;
				x=x-2;y=y-2;
				drawRect(x,y,x+w,y+h,1,STYLE_SOLID);
				w2=h/8;//number of lines
				w3=item->current-w2/2;if(w3<1) w3=1;//first item to draw
				w4=w3+w2-1;// last item to draw
				if(w4>sl->count) w4=sl->count;
				for(w5=w3;w5<=w4;w5++)
				{
				    bufp1=slistGet(sl,w5);//the label
				    bufp2=slistGet(item->values,w5);//the value
				    bufp3=slistGet(item->formats,w5);//the format
				    stringFormatValue(s,bufp3,bufp2);
				    stringSprintf(s1,"%s:  %s",bufp1,s->data);
				    PrintMini(x+w/2-GetMiniTextWidth(s1->data)/2-1,y+2+(w5-w3)*8,(unsigned char*)s1->data,MINI_OVER);
				};
				drawRect(x+2,y+1+(item->current-w3)*8,x+w-8,y+(item->current-w3+1)*8-1,1,STYLE_SOLID);
				if(w4<sl->count) PrintMini(x+w-6,y+h-6,(unsigned char*)"\xE6\x93",MINI_OVER);
				if(w5>1) PrintMini(x+w-6,y+2,(unsigned char*)"\xE6\x92",MINI_OVER);
				//LIST is the only item that gets a rectangle instead of reverse
				if(dlg->current==w1) drawRect(x+1,y+1,x+w-1,y+h-1,1,STYLE_SOLID);
			}else if(item->type==DIALOG_MENU && item->visible==1)//menu
			{
				menu=item;
				hasMenu=1;
				w2=0;
				if(w1==dlg->current) 
				{
				 jump4:
				  dlg->current++;w2++;//does not keep focus
				  if(dlg->current>n) dlg->current=1;
				  item2=dialogGet(dlg,dlg->current);
				  if((item2->type==DIALOG_MENU || item2->type==DIALOG_LABEL || item2->visible==0) && w2<dlg->count) goto jump4;
				};
				//if(w2!=dlg->current) {EVENT(dlg,EVENT_ITEM_NEXT,w2);};
				w3=stringListCount(item->text,',');
				for(w4=1;w4<=w3;w4++)
				{
					stringListItem(item->text,',',w4,s);//menu choice with submenus
					stringListItem(s,'|',1,s1);//menu choice without submenus
					w5=stringListCount(s,'|');//number of submenus +1
					if(stringLength(s1)>0) 
					{
					  MenuOption[w4-1]=1;
					  x=(w4-1)*21;
					  PrintMini(x+2+21/2-GetMiniTextWidth(s1->data)/2,55+2,(unsigned char*)s1->data,MINI_OVER);
					  if(w5>1) PrintMini(w4*21-3,58,(unsigned char*)"\x0C",MINI_OVER);
					  Bdisp_AreaReverseVRAM(x+1,55+1,x+21-1,63-1);
					};
				};
			};

		};
		EVENT(dlg,EVENT_AFTER_REFRESH,0);
		GetKey(&key);
		item=dialogGet(dlg,dlg->current);
		if(key==KEY_CTRL_DOWN && (item->type==DIALOG_LIST || item->type==DIALOG_LISTEDIT))
		{
		    //special case of LIST with KEY_CTRL_DOWN
		    w2=item->current;
		    item->current++;
		    if(item->current > item->items->count) item->current=1;
		    EVENT(dlg,EVENT_ITEM_CHANGED,w2);
		}else if(key==KEY_CTRL_DOWN || key==KEY_CTRL_RIGHT)
		{
			w3=0;
			w2=dlg->current;
		  jump1:
			dlg->current++;w3++;
			if(dlg->current>n) dlg->current=1;
			item=dialogGet(dlg,dlg->current);
			if((item->type==DIALOG_LABEL || item->type==DIALOG_MENU || item->visible==0) && w3<dlg->count) goto jump1;
			EVENT(dlg,EVENT_ITEM_NEXT,w2);
		}else if(key==KEY_CTRL_UP && (item->type==DIALOG_LIST || item->type==DIALOG_LISTEDIT))
		{
		    //special case of LIST with KEY_CTRL_UP
		    w2=item->current;
		    item->current--;
		    if(item->current < 1) item->current=item->items->count;
		    EVENT(dlg,EVENT_ITEM_CHANGED,w2);
		}
		else if(key==KEY_CTRL_UP || key==KEY_CTRL_LEFT)
		{
			w3=0;
			w2=dlg->current;
		  jump2:
			dlg->current--;
			if(dlg->current<1) dlg->current=n;
			item=dialogGet(dlg,dlg->current);
			if((item->type==DIALOG_LABEL || item->type==DIALOG_MENU || item->visible==0) && w3<dlg->count) goto jump2;
			EVENT(dlg,EVENT_ITEM_PREV,w2);
		}else if(key==KEY_CTRL_EXE)
		{
			//respond to EXE on each item
			item=dialogGet(dlg,dlg->current);
			if(item->type==DIALOG_BUTTON)// if button pressed
			{
				SetScreen(Scr1);
				free(Scr1);
				stringFree(s);
				stringFree(s1);
				dlg->retCode=dlg->current;dlg->retSubcode=0;
				return;
			} else if(item->type==DIALOG_EDITBOX)//editbox
			{
				Scr2=GetScreen();
				PopUpWin(1);
				stringInput(4,4,s,16,32,item->text->data,0,INPUT_STRING);
				stringFormatValue(s1,item->format->data,s->data);
				w2=strcmp(item->text->data,s1->data);
				stringSet(item->text,s1->data);
				SetScreen(Scr2);
				free(Scr2);
				EVENT(dlg,EVENT_ITEM_CHANGED,w2);
			} else if(item->type==DIALOG_CHECKBOX)//check box
			{
				w2=item->status;
				if(item->status==0) item->status=1; else item->status=0;
				EVENT(dlg,EVENT_ITEM_CHANGED,w2);
			} else if(item->type==DIALOG_RADIO)//radio buttons
			{
				w2=item->current;
				item->current++;
				if(item->current > item->items->count) item->current=1;
				EVENT(dlg,EVENT_ITEM_CHANGED,w2);
			} else if(item->type==DIALOG_DROPLIST) //drop down list
			{
				Scr2=GetScreen();
				w2=item->current;
				item->current=multipleChoiceBox(item->items,item->current);
				SetScreen(Scr2);
				free(Scr2);
				EVENT(dlg,EVENT_ITEM_CHANGED,w2);
			} else if(item->type==DIALOG_LIST)//list
			{
				Scr2=GetScreen();
				w2=item->current;
				item->current=multipleChoiceBox(item->items,item->current);
				SetScreen(Scr2);
				free(Scr2);
				EVENT(dlg,EVENT_ITEM_CHANGED,w2);
			} else if(item->type==DIALOG_LISTEDIT)//listedit
			{
				bufp2=slistGet(item->values,item->current);//the value
				bufp3=slistGet(item->formats,item->current);//the format
				Scr2=GetScreen();
				PopUpWin(1);
				stringInput(4,4,s,16,32,bufp2,0,INPUT_STRING);
				if(item->status==0)//if not Readonly
				{
				  stringFormatValue(s1,bufp3,s->data);//s1 holds the new value
				  w2=strcmp(bufp2,s1->data);
				  slistSet(item->values,item->current,s1->data);
				  EVENT(dlg,EVENT_ITEM_CHANGED,w2);
				};
				SetScreen(Scr2);
				free(Scr2);
			};
		} else if(key>=KEY_CTRL_F1 && key<=KEY_CTRL_F6 && hasMenu==1)
		{
			w2=key-KEY_CTRL_F1+1;
			if(menu->visible==1 && MenuOption[w2-1]==1) 
			{
			  stringListItem(menu->text,',',w2,s);//the menu
			  w5=stringListCount(s,'|')-1;//num of submenus
			  if(w5>0)//has submenus
			  {
			    w3=stringIdx(s,'|',1);
			    stringMid(s,w3+1,stringLength(s),s1);//the submenus in s1
			    w4=execMenuPopup(s1,w2);
			    if(w4>0)
			    {
			      SetScreen(Scr1);
			      free(Scr1);
			      stringFree(s);
			      stringFree(s1);
			      dlg->retCode=100;
			      dlg->retSubcode=w2*100+w4;
			      return;
			    };
			  }else
			  {
			    SetScreen(Scr1);
			    free(Scr1);
			    stringFree(s);
			    stringFree(s1);
			    dlg->retCode=100;
			    dlg->retSubcode=w2*100;
			    return;
			  }
			};//F1-F6 selected
		} else //all other key presses
		{
		  EVENT(dlg,EVENT_KEY_PRESS,key);
		};
	};
	SetScreen(Scr1);
	free(Scr1);
	stringFree(s);
	stringFree(s1);
	return;

}




struct StringListNode* slistCreateNode(const char* text)
{
  struct StringListNode* item;
  int len;
  item=malloc(sizeof(struct StringListNode));
  ASSERT(item!=0);
  len=strlen(text)+1;
  item->data=(char *)malloc(sizeof(char)*len);
  ASSERT(item->data!=0);
  item->maxCount=len;
  item->next=0;
  strcpy(item->data,text);
  return item;
}

void slistPrepend(struct StringList* list,const char* text)
{
  struct StringListNode* item;
  struct StringListNode* tmp;
  item=slistCreateNode(text);
  tmp=list->first;
  list->first=item;
  item->next=tmp; 
  list->count++;
}


struct StringList* slistInit(int n,int itemsize)
{
  struct StringList* list;
  struct StringListNode* item;
  char* bufp;
  int w1;
  list=malloc(sizeof(struct StringList));
  ASSERT(list!=0);
  list->first=0;
  list->count=0;
  list->current=0;
  bufp=(char *)malloc(sizeof(char)*(itemsize+1));
  ASSERT(bufp!=0);
  memset(bufp,32,itemsize);//fill with space(ascii 32)
  bufp[itemsize]=0;//0 to end
  for(w1=1;w1<=n;w1++)
  {
    slistPrepend(list,bufp);
    list->first->data[0]=0;//0 to end
  };
  free(bufp);
  return list;
}

struct StringListNode* slistFirst(struct StringList* list)
{
  list->current=list->first;
  return list->current;
}

struct StringListNode* slistNext(struct StringList* list)
{
  if(list->current!=0)
  {
    list->current=list->current->next;
  };
  return list->current;
}

char* slistGet(struct StringList* l,int i)
{
  struct StringListNode* item;
  int j;
  j=1;
  item=l->first;
  ASSERT(item!=0);
  do {
    if(j==i) return(item->data);
    item=item->next;
    j++;
  } while(item!=0);
  ASSERT(-1);//to provoke an assert if there is no item
  return 0;
}

void slistSetAll(struct StringList* l, char d, const char* items)
{
  int w1,w2,n;
  struct String* s;
  struct String* s1;
  s=stringInit(items,1);
  s1=stringInit("",32);
  slistRemoveAll(l);
  n=stringListCount(s,d);
  for(w1=1;w1<=n;w1++)
  {
    stringListItem(s,d,w1,s1);
    slistAppend(l,s1->data);
  };
  stringFree(s1);
  stringFree(s);
}

void slistGetAll(struct StringList* l, char d, struct String* items)
{
  int w1,n;
  struct StringListNode* item;
  char buf[2];
  buf[0]=d;buf[1]=0;
  stringSet(items,"");
  item=l->first;
  while(item!=0)
  {
    stringAppend(items,item->data);
    if(item->next!=0) stringAppend(items,buf);
    item=item->next;
  };
}

void slistCopy(struct StringList* dest, const struct StringList* src)
{
  struct StringListNode* item;
  slistRemoveAll(dest);
  item=src->first;
  while(item!=0)
  {
    slistAppend(dest,item->data);
    item=item->next;
  };
}

void slistNodeSet(struct StringListNode* item, const char* text)
{
  int len;
  ASSERT(item!=0);
  len=strlen(text)+1;
  if(len>(item->maxCount))//needs realloc
  {
    free(item->data);
    item->data=(char *)malloc(sizeof(char)*len);
    ASSERT(item->data);
    item->maxCount=len;
  };
  strcpy(item->data,text);
  return;
}


void slistSet(struct StringList* l,int i,const char* text)
{
  struct StringListNode* item;
  int j,len;
  j=1;
  item=l->first;
  ASSERT(item!=0);
  do {
    if(j==i)
    {
      slistNodeSet(item,text);
    };
    item=item->next;
    j++;
  } while(item!=0);
  ASSERT(-1);// to provoke an assert if item not exist
}

void slistRemoveAll(struct StringList* l)
{
  struct StringListNode* item;
  struct StringListNode* tmp;
  item=l->first;
  while(item!=0)
  {
    free(item->data);
    tmp=item;
    item=item->next;
    free(tmp);
  };
  l->count=0;
  l->current=0;
  l->first=0;
}

void slistFree(struct StringList* l)
{
  struct StringListNode* item;
  struct StringListNode* tmp;
  item=l->first;
  while(item!=0)
  {
    free(item->data);
    tmp=item;
    item=item->next;
    free(tmp);
  };
  free(l);
}

void slistInsertAt(struct StringList* l,int i,const char* text)
{
  struct StringListNode* item;
  struct StringListNode* prev;
  struct StringListNode* newi;
  int j;
  item=l->first;
  prev=0;
  j=1;
  do {
    if(i==j)//insert here
    {
      if(prev==0)
      {
	slistPrepend(l,text);
      }else
      {
	newi=slistCreateNode(text);
	prev->next=newi;
	newi->next=item;
	l->count++;
      };
      return;
    };
    prev=item;
    item=item->next;
    j++;
  } while(item!=0);
  ASSERT(-1); //to provoke assert if item position is not found
  return;
}

void slistAppend(struct StringList* l,const char* text)
{
  struct StringListNode* item;
  struct StringListNode* newi;
  item=l->first;
  if(item!=0)
  {
   while(item->next!=0)
   {
     item=item->next;
   };
   newi=slistCreateNode(text);
   item->next=newi;
   l->count++;
  }else //list is empty
  {
    newi=slistCreateNode(text);
    l->first=newi;
    l->count++; 
  };
}

void slistRemoveAt(struct StringList* l,int i)
{
  struct StringListNode* item;
  struct StringListNode* prev;
  struct StringListNode* newi;
  int j;
  item=l->first;
  prev=0;
  j=1;
  do {
    if(i==j)//remove here
    {
      if(prev==0)
      {
	l->first=item->next;
	free(item->data);
	free(item);
      }else
      {
	prev->next=item->next;
	free(item->data);
	free(item);
      };
      l->count--;
      return;
    };
    prev=item;
    item=item->next;
    j++;
  } while(item!=0);
  return;
}




struct Matrix* matrixInit(struct Matrix* mat,int rows,int cols)
{
	int size=rows*cols;
	int w1;
	ASSERT(rows>0);
	ASSERT(cols>0);
	mat=malloc(sizeof(struct Matrix));
	ASSERT(mat!=0);
	mat->rows=rows;
	mat->cols=cols;
	mat->data=malloc(sizeof(double)*size);
	ASSERT((mat->data)!=0);
	for(w1=0;w1<size;w1++) mat->data[w1]=0.0;
	return mat;
}

void matrixResize(struct Matrix* mat, int rows, int cols)
{
	matrixFree(mat);
	matrixInit(mat,rows,cols);
}


double matrixGet(struct Matrix* mat,int r,int c)
{
	ASSERT((r>0) && (r<=(mat->rows)));
	ASSERT((c>0) && (c<=(mat->cols)));
	return(mat->data[(r-1)*mat->cols+c-1]);
}

void matrixSet(struct Matrix* mat,int r,int c,double v)
{
	ASSERT((r>0) && (r<=(mat->rows)));
	ASSERT((c>0) && (c<=(mat->cols)));
	mat->data[(r-1)*mat->cols+c-1]=v;
}

void matrixSetAll(struct Matrix* mat,double** values)
{
  int s;
  s=mat->rows*mat->cols*sizeof(double);
  memcpy(mat->data,values,s);
}


void matrixFree(struct Matrix* mat)
{
	free(mat->data);
	free(mat);
}

void matrixZero(struct Matrix* mat)
{
	int w1,n;
	n=mat->rows * mat->cols;
	for(w1=0;w1<n;w1++)
	{
		mat->data[w1]=0.0;
	};
}

struct Matrix* matrixAppendRow(struct Matrix* mat)
{
	struct Matrix* newmat;
	int w1,w2;
	newmat=matrixInit(newmat,mat->rows+1,mat->cols);
	for(w1=1;w1<=mat->rows;w1++)
	{
		for(w2=1;w2<=mat->cols;w2++)
		{
			matrixSet(newmat,w1,w2, matrixGet(mat,w1,w2));
		};
	};
	for(w2=1;w2<=mat->cols;w2++)
	{
		matrixSet(newmat,mat->rows+1,w2,0.0);
	};
	return(newmat);
}

struct Matrix* matrixRemoveRow(struct Matrix* mat, int r)
{
	struct Matrix* newmat;
	int w1,w2,w3,w4;
	if(mat->rows==1) return(0);
	newmat=matrixInit(newmat,mat->rows-1,mat->cols);
	w3=0;w4=0;
	for(w1=1;w1<=mat->rows;w1++)
	{
		if(w1!=r)
		{
			w3++;w4++;
			for(w2=1;w2<=mat->cols;w2++)
			{
				matrixSet(newmat,w3,w2, matrixGet(mat,w4,w2));
			};
		}else w4++;
	};
	return(newmat);
}

struct Matrix* matrixRemoveColumn(struct Matrix* mat,int c)
{
	struct Matrix* newmat;
	int w1,w2,w3,w4;
	if(mat->cols==1) return (0);
	newmat=matrixInit(newmat,mat->rows,mat->cols-1);
	w3=0;w4=0;
	for(w1=1;w1<=mat->cols;w1++)
	{
		if(w1!=c)
		{
			w3++;w4++;
			for(w2=1;w2<=mat->rows;w2++)
			{
				matrixSet(newmat,w2,w3,matrixGet(mat,w2,w4));
			};
		}else w4++;
	};
	return(newmat);
}

void matrixAddMatrix(struct Matrix* mat1,struct Matrix* mat2)
{
	int w1,w2;
	for(w1=1;w1<=mat1->rows;w1++)
	{
		for(w2=1;w2<=mat1->cols;w2++)
		{
			matrixSet(mat1,w1,w2,matrixGet(mat1,w1,w2)+matrixGet(mat2,w1,w2));
		};
	};
}

void matrixAddNumber(struct Matrix* mat1,double v)
{
	int w1,w2;
	for(w1=1;w1<=mat1->rows;w1++)
	{
		for(w2=1;w2<=mat1->cols;w2++)
		{
			matrixSet(mat1,w1,w2,matrixGet(mat1,w1,w2)+v);
		};
	};
}

struct Matrix* matrixMulMatrix(struct Matrix* mat1,struct Matrix* mat2)
{
	int w1,w2,w3;
	double d1;
	struct Matrix* mat3;
	if(mat1->cols!=mat2->rows) return 0;
	mat3=matrixInit(mat3,mat1->rows,mat2->cols);
	for(w1=1;w1<=mat1->rows;w1++) // running through the rows of the new matrix
	{
		for(w2=1;w2<=mat2->cols;w2++) //running through the cols of the new matrix
		{
			d1=0.0;
			for(w3=1;w3<=mat1->cols;w3++)
			{
				d1=d1+matrixGet(mat1,w1,w3)*matrixGet(mat2,w3,w2);
			};
			matrixSet(mat3,w1,w2,d1);
		};
	};
	return mat3;
}
struct Matrix* matrixCopy(struct Matrix* mat1)
{
	int w1;
	struct Matrix* newmat;
	if(mat1==0) return 0;
	newmat=matrixInit(newmat,mat1->rows,mat1->cols);
	for(w1=0;w1<mat1->rows*mat1->cols;w1++)
	{
		newmat->data[w1]=mat1->data[w1];
	};
	return newmat;
}

struct Matrix* matrixTranspose(struct Matrix* mat1)
{
	int w1,w2;
	struct Matrix* newmat;
	newmat=matrixInit(newmat,mat1->cols,mat1->rows);
	for(w1=1;w1<=mat1->rows;w1++)
	{
		for(w2=1;w2<=mat1->cols;w2++)
		{
			matrixSet(newmat,w2,w1, matrixGet(mat1,w1,w2));
		};
	};
	return newmat;
}

void matrixMulNumber(struct Matrix* mat1,double v)
{
	int w1,w2;
	for(w1=1;w1<=mat1->rows;w1++)
	{
		for(w2=1;w2<=mat1->cols;w2++)
		{
			matrixSet(mat1,w1,w2,matrixGet(mat1,w1,w2)*v);
		};
	};
}

void matrixSwapRow(struct Matrix* mat1,int r1,int r2)
{
	int w1;
	double d1;
	for(w1=1;w1<=mat1->cols;w1++)
	{
		d1=matrixGet(mat1,r1,w1);
		matrixSet(mat1,r1,w1,matrixGet(mat1,r2,w1));
		matrixSet(mat1,r2,w1,d1);
	};
}

void matrixAddMultiplyRow(struct Matrix* mat1,int r1,int r2,double v)
{
	int w1;
	double d1;
	for(w1=1;w1<=mat1->cols;w1++)
	{
		d1=matrixGet(mat1,r1,w1)+matrixGet(mat1,r2,w1)*v;
		matrixSet(mat1,r1,w1,d1);
	};
}

void matrixDisplay(struct Matrix* a)//TODO OBSOLETE
{
	char buf1[32];
	int w1,w2;
	unsigned int key;
	unsigned char* Scr1;
	Scr1=GetScreen();
	Bdisp_AllClr_VRAM();
	for(w1=1;w1<=a->rows;w1++)
	{
		for(w2=1;w2<=a->cols;w2++)
		{
			locate(1+(w2-1)*5,w1);
			sprintf(buf1,"%.2f",matrixGet(a,w1,w2));
			Print((unsigned char *)buf1);
		};
	};
	GetKey(&key);
	SetScreen(Scr1);
	free(Scr1);

}

struct Matrix* matrixConjGradient(struct Matrix* a,struct Matrix* b,double residual)
{
	int w1,w2,w3,w4;
	double d1,d2;
	int n,m;
	double ak,bk,res;
	struct Matrix* matRK;
	struct Matrix* matRK1;
	struct Matrix* matXK;
	struct Matrix* matPK;
	struct Matrix* mat0;
	struct Matrix* mat1;
	struct Matrix* mat2;
	if(a->rows!=a->cols) return 0;
	if(a->rows!=b->rows) return 0;
	n=a->rows;
	m=b->cols;
	if(m!=1) return 0;
	matRK=matrixInit(matRK,n,1);
	matRK1=matrixInit(matRK1,n,1);
	matXK=matrixInit(matXK,n,1);
	matPK=matrixInit(matPK,n,1);
	//set XK=X0 = {1,1,...}
	for(w1=1;w1<=n;w1++)
	{
		matrixSet(matXK,w1,1,1.0);
	};
	//set RK=R0=B-A*X0
	mat0=matrixMulMatrix(a,matXK);
	matrixMulNumber(mat0,-1.0);
	matrixAddMatrix(mat0,b);
	//set PK=P0=R0
	for(w1=1;w1<=n;w1++)
	{
		matrixSet(matPK,w1,1,matrixGet(mat0,w1,1));
		matrixSet(matRK,w1,1,matrixGet(mat0,w1,1));
	};
	matrixFree(mat0);
	//start loop
	res=1.0E+10+residual;
	while(res>residual)
	{
		//ak=rk^T rk /(pk^T A pk)
		mat0=matrixTranspose(matRK);
		mat1=matrixMulMatrix(mat0,matRK);//1x1 matrix
		d1=matrixGet(mat1,1,1);
		matrixFree(mat0);matrixFree(mat1);
		mat0=matrixTranspose(matPK);
		mat1=matrixMulMatrix(mat0,a);
		mat2=matrixMulMatrix(mat1,matPK);//1x1 matrix
		d2=matrixGet(mat2,1,1);
		matrixFree(mat0);matrixFree(mat1);matrixFree(mat2);
		ak=d1/d2;
		//xk+1=xk+ak pk
		for(w1=1;w1<=n;w1++)
		{
			matrixSet(matXK,w1,1,matrixGet(matXK,w1,1)+ak*matrixGet(matPK,w1,1));
		};
		//rk+1=rk-ak A pk
		mat0=matrixMulMatrix(a,matPK);
		for(w1=1;w1<=n;w1++)
		{
			matrixSet(matRK1,w1,1,matrixGet(matRK,w1,1)-ak*matrixGet(mat0,w1,1));
		};
		matrixFree(mat0);
		//compute residual res
		res=0.0;
		for(w1=1;w1<=n;w1++)
		{
			res=res+fabs(matrixGet(matRK1,w1,1));
		};
		//bk=rk+1^T rk+1 / (rk^T rk)
		mat0=matrixTranspose(matRK1);
		mat1=matrixMulMatrix(mat0,matRK1);//1x1 matrix
		d1=matrixGet(mat1,1,1);
		matrixFree(mat0);matrixFree(mat1);
		mat0=matrixTranspose(matRK);
		mat1=matrixMulMatrix(mat0,matRK);//1x1 matrix
		d2=matrixGet(mat1,1,1);
		matrixFree(mat0);matrixFree(mat1);
		bk=d1/d2;
		//pk+1=rk+1 + bk pk
		for(w1=1;w1<=n;w1++)
		{
			matrixSet(matPK,w1,1,matrixGet(matRK1,w1,1)+bk*matrixGet(matPK,w1,1));
		};
		//matRK=matRK1 (k=k+1)
		for(w1=1;w1<=n;w1++)
		{
			matrixSet(matRK,w1,1,matrixGet(matRK1,w1,1));
		};
	};
	//copy XK solution to B
	for(w1=1;w1<=n;w1++)
	{
		matrixSet(b,w1,1,matrixGet(matXK,w1,1));
	};
	matrixFree(matPK);
	matrixFree(matXK);
	matrixFree(matRK);

	return b;
}

struct Matrix* matrixLinearSolve(struct Matrix* a,struct Matrix* b)
{
	int w1,w2,w3,w4,w5;
	int n,m;
	double piv,cl;

	if(a->rows!=a->cols) return 0;//non square - error
	if(a->rows!=b->rows) return 0;//a not same rows with b
	n=a->rows;
	m=b->cols;
	for(w1=1;w1<=(n-1);w1++) // rows 1 to n-1
	{
		piv=matrixGet(a,w1,w1);
		if(piv==0.0) return 0; // zero pivot
		w2=w1+1;
		for(w3=w2;w3<=n;w3++)
		{
			cl=matrixGet(a,w3,w1)/piv;
			if(cl!=0.0)
			{
				for(w4=1;w4<=m;w4++)
				{
					matrixSet(b,w3,w4, matrixGet(b,w3,w4)-cl*matrixGet(b,w1,w4));
				};
				for(w4=w2;w4<=n;w4++)
				{
					matrixSet(a,w3,w4, matrixGet(a,w3,w4)-cl*matrixGet(a,w1,w4));
				};
			};
		};
	};
	//now backsubstitution
	for(w4=1;w4<=m;w4++)
	{
		if(matrixGet(a,n,n)==0.0) return 0; //zero div
		matrixSet(b,n,w4, matrixGet(b,n,w4)/matrixGet(a,n,n));
	};
	for(w5=1;w5<=m;w5++)
	{
		for(w2=n-1;w2>=1;w2--)
		{
			cl=0.0;
			for(w4=w2+1;w4<=n;w4++)
			{
				cl=cl+matrixGet(a,w2,w4)*matrixGet(b,w4,w5);
			};
			if(matrixGet(a,w2,w2)==0.0) return 0; //zero div
			matrixSet(b,w2,w5, (matrixGet(b,w2,w5)-cl)/matrixGet(a,w2,w2));
		};
	};
	return(b);
}


double matrixDeterminant(struct Matrix* a)
{
	struct Matrix* m1;
	struct Matrix* m2;
	int n,w1;
	double det=0;
	signed int sign1;
	if(a->rows!=a->cols) return 0.0;
	n=a->rows;
	if(n==1) {det=matrixGet(a,1,1);return det;};
	if(n==2)
	{
		det=matrixGet(a,1,1)*matrixGet(a,2,2)-matrixGet(a,1,2)*matrixGet(a,2,1);
		return det;
	};
	if(n==3) 
	{
		det=matrixGet(a,1,1)*matrixGet(a,2,2)*matrixGet(a,3,3)
			+matrixGet(a,1,2)*matrixGet(a,2,3)*matrixGet(a,3,1)
			+matrixGet(a,1,3)*matrixGet(a,2,1)*matrixGet(a,3,2)
			-matrixGet(a,1,3)*matrixGet(a,2,2)*matrixGet(a,3,1)
			-matrixGet(a,1,2)*matrixGet(a,2,1)*matrixGet(a,3,3)
			-matrixGet(a,1,1)*matrixGet(a,2,3)*matrixGet(a,3,2);
		return det;
	};
	sign1=+1;
	for(w1=1;w1<=n;w1++)
	{
		m1=matrixRemoveRow(a,w1);
		m2=matrixRemoveColumn(m1,1);
		matrixFree(m1);
		det=det+sign1*matrixGet(a,w1,1)*matrixDeterminant(m2);
		matrixFree(m2);
		if(sign1==+1) sign1=-1; else sign1==+1;
	};
	return det;
}


int execMenu(const char* items) 
{
  struct Dialog* dlg;
  struct DialogItem* item;
  int r;
  dlg=dialogInit();
  item=dialogItemInit();
  dialogItemSetMenu(item,items);
  dialogAppend(dlg,item);
  dialogExec(dlg,"",0);
  r=dlg->retSubcode;
  dialogItemFree(item);
  dialogFree(dlg);
  return r;
}

int execMenuPopup(struct String* items,int pos)
{
  //items are separated with |
  unsigned char* Scr1;
  unsigned int key;
  struct String* s;
  int pxsize=24;//21+3
  int i,x,y,current;
  int n=stringListCount(items,'|');
  s=stringInit(items->data,1);
  for(i=1;i<=n;i++) 
  {
    stringListItem(items,'|',i,s);
    pxsize=max(pxsize,GetMiniTextWidth(s->data)+3);
  };
  Scr1=GetScreen();
  current=n;
  while(1){
  SetScreen(Scr1);
  if(pos!=6)
  {
    fillRect((pos-1)*21+1,55-n*7-1,(pos-1)*21+pxsize,55,0);
    drawRect((pos-1)*21+1,55-n*7-1,(pos-1)*21+pxsize,55,1,STYLE_SOLID);
  }else
  {
    fillRect(127-pxsize,55-n*7-1,127,55,0);
    drawRect(127-pxsize,55-n*7-1,127,55,1,STYLE_SOLID);
  };
  for(i=1;i<=n;i++)
  {
	stringListItem(items,'|',i,s);
	if(pos!=6) x=(pos-1)*21+2+pxsize/2-GetMiniTextWidth(s->data)/2; else x=127-pxsize+2;
	y=55-i*7+1;
	PrintMini(x,y,(unsigned char *)s->data,MINI_OVER);
  };
  if(pos!=6) Bdisp_AreaReverseVRAM((pos-1)*21+2,55-current*7,(pos-1)*21+pxsize-1,55-(current-1)*7-1);
    else Bdisp_AreaReverseVRAM(127-pxsize+1,55-current*7,126,55-(current-1)*7-1);
  GetKey(&key);
  if(key==KEY_CTRL_UP)
  {
    current++;
    if(current>n) current=1;
  } else if(key==KEY_CTRL_DOWN)
  {
    current--;
    if(current<1) current=n;
  }else if(key==KEY_CTRL_EXIT || key==KEY_CTRL_LEFT || key==KEY_CTRL_RIGHT)
  {
    current=0;
    break;
  }else if(key==KEY_CTRL_EXE)
  {
	break;
  };
  };//while
  SetScreen(Scr1);
  free(Scr1);
  stringFree(s);
  return current;
}


void drawCircle(int x0,int y0,int r,int color)
{
  int f = 1 -r;
  int ddF_x = 1;
  int ddF_y = -2 * r;
  int x = 0;
  int y = r;
  Bdisp_SetPoint_VRAM(x0,y0+r,color);
  Bdisp_SetPoint_VRAM(x0,y0-r,color);
  Bdisp_SetPoint_VRAM(x0+r,y0,color);
  Bdisp_SetPoint_VRAM(x0-r,y0,color);
  while(x < y)
  {
    if(f >= 0)
    {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
    Bdisp_SetPoint_VRAM(x0 + x, y0 + y,color);
    Bdisp_SetPoint_VRAM(x0 - x, y0 + y,color);
    Bdisp_SetPoint_VRAM(x0 + x, y0 - y,color);
    Bdisp_SetPoint_VRAM(x0 - x, y0 - y,color);
    Bdisp_SetPoint_VRAM(x0 + y, y0 + x,color);
    Bdisp_SetPoint_VRAM(x0 - y, y0 + x,color);
    Bdisp_SetPoint_VRAM(x0 + y, y0 - x,color);
    Bdisp_SetPoint_VRAM(x0 - y, y0 - x,color);
  }
}


/*void drawLine(int x1, int y1, int x2, int y2, int color)
{
	if(color==1)
		Bdisp_DrawLineVRAM(x1,y1,x2,y2);
		else if(color==0) Bdisp_ClearLineVRAM(x1,y1,x2,y2);
}*/

void drawLine(int x1, int y1, int x2, int y2, int color,int pattern)
{
  int dx,dy,sx,sy,err,e2,pi,c;
  pi=1;
  dx=abs(x2-x1);
  dy=abs(y2-y1);
  if(x1<x2) sx=1; else sx=-1;
  if(y1<y2) sy=1; else sy=-1;
  err=dx-dy;
  while(1)
  {
    c=pattern & pi;
    if(color==0) if(c==0) c=1; else c=0;
    Bdisp_SetPoint_VRAM(x1,y1,c);
    if(x1==x2 && y1==y2) break;
    e2=2*err;
    if(e2>-dy)
    {
      err=err-dy;
      x1=x1+sx;
    };
    if(e2<dx)
    {
      err=err+dx;
      y1=y1+sy;
    };
    pi=pi<<1; if(pi>128) pi=1;
  }; 
}

void drawRect(int x1,int y1,int x2,int y2,int color,int pattern)
{
  drawLine(x1,y1,x2,y1,color,pattern);
  drawLine(x2,y1,x2,y2,color,pattern);
  drawLine(x2,y2,x1,y2,color,pattern);
  drawLine(x1,y2,x1,y1,color,pattern);
}

void fillRect(int x1,int y1,int x2,int y2, int color)
{
  DISPBOX b;
  b.left=min(x1,x2);b.right=max(x1,x2);
  b.top=min(y1,y2);b.bottom=max(y1,y2);
  Bdisp_AreaClr_VRAM(&b);
  if(color!=0) 
  {
    Bdisp_AreaReverseVRAM(x1,y1,x2,y2);
  };
}


void viewSet(double leftx, double topy, double rightx, double bottomy, struct View* v)
{
  v->leftX=leftx;
  v->topY=topy;
  v->rightX=rightx;
  v->bottomY=bottomy;
  v->dx=v->rightX - v->leftX;
  v->dy=v->bottomY - v->topY;
  ASSERT(v->dx!=0);ASSERT(v->dy!=0);
  v->pixeldX=1.0/128.0*v->dx;
  v->pixeldY=1.0/64.0*v->dy;
}


void viewSetOrigin(int x0, int y0, double dx, double dy,struct View* v)
{
  v->leftX=-x0/128.0*dx;
  v->topY=-y0/64.0*dy;
  v->rightX=(128.0-x0)/128.0*dx;
  v->bottomY=(64.0-y0)/64.0*dy;
  v->dx=v->rightX - v->leftX;
  v->dy=v->bottomY - v->topY;
  ASSERT(v->dx!=0);ASSERT(v->dy!=0);
  v->pixeldX=1.0/128.0*v->dx;
  v->pixeldY=1.0/64.0*v->dy;
}

void viewPointToScreen(double x, double y, int* xi, int* yi, struct View* v)
{
  *xi=floor(128.0*(x-v->leftX)/v->dx);
  *yi=floor(64.0*(y-v->topY)/v->dy);
}

void viewPointToView(int xi, int yi, double* x, double* y, struct View* v)
{
 *x=(v->dx*xi)/128.0+v->leftX;
 *y=(v->dy*yi)/64.0+v->topY;
}


void viewPoint(double x, double y, int color, struct View* v)
{
  int xi,yi;
  xi=floor(128.0*(x-v->leftX)/v->dx);
  yi=floor(64.0*(y-v->topY)/v->dy);
  Bdisp_SetPoint_VRAM(xi,yi,color);
}


void viewLine(double x1, double y1, double x2, double y2, int color, int pattern, struct View* v)
{
	int i1,j1,i2,j2;
	i1=floor(128*(x1 - v->leftX)/v->dx);
	j1=floor(64.0*(y1 - v->topY)/v->dy);
	i2=floor(128.0*(x2 - v->leftX)/v->dx);
	j2=floor(64.0*(y2 - v->topY)/v->dy);
	drawLine(i1,j1,i2,j2,color,pattern);
}

void viewRect(double x1, double y1, double x2, double y2, int color, int pattern, struct View* v)
{
	int i1,j1,i2,j2;
	i1=floor(128*(x1 - v->leftX)/v->dx);
	j1=floor(64.0*(y1 - v->topY)/v->dy);
	i2=floor(128.0*(x2 - v->leftX)/v->dx);
	j2=floor(64.0*(y2 - v->topY)/v->dy);
	drawRect(i1,j1,i2,j2,color,pattern);
}

void viewFillRect(double x1, double y1, double x2, double y2,int color, struct View* v)
{
	int i1,j1,i2,j2;
	i1=floor(128*(x1 - v->leftX)/v->dx);
	j1=floor(64.0*(y1 - v->topY)/v->dy);
	i2=floor(128.0*(x2 - v->leftX)/v->dx);
	j2=floor(64.0*(y2 - v->topY)/v->dy);
	fillRect(i1,j1,i2,j2,color);
}


void viewScale(double x0, double y0, double scale, struct View* v)
{
	int x1,y1;
	x1=floor(128.0*(x0-v->leftX)/v->dx);
	y1=floor(64.0*(y0-v->topY)/v->dy);
	viewSet(x1,y1,scale*v->dx,scale*v->dy,v);
}


void viewPrintMini(double x, double y, const char* str, int type, int align,struct View* v)
{
	int i,j,d;
	i=floor(128*(x - v->leftX)/v->dx);
	j=floor(64.0*(y - v->topY)/v->dy);
	if(align & ALIGN_CENTER)
	{
	  d=GetMiniTextWidth(str);
	  i=i-d/2;if(i<0 && i>=-d/2) i=0;
	}else if(align & ALIGN_RIGHT)
	{
	  d=GetMiniTextWidth(str);
	  i=i-d;if(i<0 && i>=-d) i=0;
	};
	if(align & ALIGN_VERTICAL)
	{
	  j=j-2;if(j<0 && j>=-2) j=0;
	};
	PrintMini(i,j,(unsigned char*)str,type);
}

void viewArrow(double x1, double y1, double x2, double y2, int size, int color, int pattern, struct View* v)
{
  double th,ar,dx,dy,d,x3,y3,x4,y4;
  th=lineAngle(x1,y1,x2,y2);
  ar=PI/5;
  d=max(v->pixeldX,v->pixeldY)*size;
  dx=cos(ar)*d;dy=sin(ar)*d;
  x3=x2-cos(th)*dx-sin(th)*dy;
  y3=y2-sin(th)*dx+cos(th)*dy;
  x4=x2-cos(th)*dx+sin(th)*dy;
  y4=y2-sin(th)*dx-cos(th)*dy;
  viewLine(x1,y1,x2,y2,color,pattern,v);
  viewLine(x2,y2,x3,y3,color,pattern,v);
  viewLine(x2,y2,x4,y4,color,pattern,v);
}

double lineAngle(double x1,double y1,double x2,double y2)
{
	double dx,dy;
	double d1=0.0;
	dx=x2-x1;dy=y2-y1;
	if(dx==0.0 && dy<0.0) d1=3.0*PI/2.0;
		else if(dx==0.0 && dy>0.0) d1=PI/2.0;
		else if(dx>0.0 && dy>0.0) d1=atan(dy/dx);
		else if(dx>0.0 && dy<0.0) d1=2.0*PI+atan(dy/dx);
		else if(dx<0.0) d1=PI+atan(dy/dx);
	return d1;
}

/*
void viewPerpendicularLine(double x1, double y1, double x2, double y2, double x, double y, struct View* v)
{
	double f,t;
	double d1,d2;
	double d3,d4;
	t=lineAngle(x1,y1,x2,y2);
	d1=x1+x*cos(t);
	d2=y1+x*sin(t);
	d3=-y*sin(t);
	d4=+y*cos(t);
	viewLine(d1,d2,d1+d3,d2+d4,1,v);
}*/

int isNumberKey(unsigned int key)
{
	if(key>=KEY_CHAR_0 && key<=KEY_CHAR_9) return key;
	if(key==KEY_CHAR_EXP) return 69;//char 'E'
	if(key==KEY_CHAR_DP) return 46;//char '.'
	if(key==KEY_CHAR_PMINUS || key==KEY_CHAR_MINUS) return 45;//char '-'
	return 0;
}

int isAlphaKey(unsigned int key)
{
	if(key>=KEY_CHAR_A && key<=KEY_CHAR_Z) return key;
	if(key==KEY_CHAR_SPACE || key==KEY_CHAR_VALR || key==KEY_CHAR_THETA || key==KEY_CHAR_DQUATE) return key;
	return 0;
}

void stringInput(int colx, int rowy, struct String* s, int len, int maxLen, const char* preinput, int hasBox, int type) 
{
	char b[2];
	//char btemp[1];
	//char bclear[22];
	int pos=1;//current pos 
	int pos0=1;//start pos
	int w1;
	unsigned int key;
	struct String* s1;
	s1=stringInit("",maxLen);
	b[1]='\0';
	if(preinput!=0) if(strlen(preinput)>0)
	{
	  stringSet(s,preinput);
	  pos=stringLength(s)+1;
	  if(pos<1) pos=1;
	  if(pos>maxLen) pos=maxLen;
	  if(pos>=len) pos0=pos-len+1;
	} else stringSet(s,"");
	while(1)
	{
		if(hasBox==1) drawRect((colx-1)*6-1,(rowy-1)*8-2,(colx+len-1)*6+1,rowy*8,1,STYLE_SOLID);
		fillRect((colx-1)*6,(rowy-1)*8-1,(colx+len-1)*6,rowy*8-1,0);//clear
		locate(colx,rowy);
		stringMid(s,pos0,len,s1);
		Print((unsigned char *)s1->data);
		drawLine((colx+pos-pos0-1)*6+1,rowy*8-1,(colx+pos-pos0-1)*6+5,rowy*8-1,1,STYLE_SOLID);
		drawLine((colx+pos-pos0-1)*6+1,rowy*8-2,(colx+pos-pos0-1)*6+5,rowy*8-2,1,STYLE_SOLID);
		if(pos0>1) drawLine((colx-1)*6,(rowy-1)*8-1,(colx-1)*6,rowy*8-1,1,STYLE_DOT);
		if(pos0+len-1<stringLength(s)) drawLine((colx+len-1)*6,(rowy-1)*8-1,(colx+len-1)*6,rowy*8-1,1,STYLE_DOT);
		b[0]='\1';
		GetKey(&key);
		if(key==KEY_CTRL_EXE) break;
		if(key==KEY_CTRL_AC)
		{
		  pos=1;pos0=1;
		  stringSet(s,"");
		}else if (key==KEY_CTRL_DEL && stringLength(s)>0)
		{
		  if(pos>stringLength(s) || pos==maxLen)//remove last char
		  {
		    s->data[stringLength(s)-1]='\0';
		    pos--;
		  }else if(pos<=stringLength(s))//remove current pos char
		  {
		    stringRemoveAt(s,pos,1);
		  };
		}else if (key==KEY_CTRL_LEFT)
		{
		  pos--;
		}else if (key==KEY_CTRL_RIGHT)
		{
		  pos++;
		}else if (key==KEY_CTRL_UP)
		{
		  pos=1;
		}else if (key==KEY_CTRL_DOWN)
		{
		  pos=stringLength(s)+1;
		};
		if(type==INPUT_NUMBER) 
		{
		  w1=isNumberKey(key);
		  if(w1!=0) b[0]=(char)w1;
		}else if(type==INPUT_STRING)
		{
		  w1=isAlphaKey(key);
		  if(w1!=0) b[0]=(char)w1; 
		    else 
		    {
		      w1=isNumberKey(key);
		      if(w1!=0) b[0]=(char)w1;
		    };
		};
		if(b[0]!='\1')//if key is valid
		{
		  if(pos>stringLength(s))
		  {
		    stringAppend(s,(char*)&b);//append last char
		  }else 
		  {
		    s->data[pos-1]=b[0];//replace pos char
		  };
		  pos++;
		};
		if(pos<1) pos=1;
		if(pos>stringLength(s)+1) pos=stringLength(s)+1;
		if(pos>maxLen) pos=maxLen;
		pos0=1;
		if(pos>=len) pos0=pos-len+1;
	};
  stringFree(s1);
}


/*
//TODO OBSOLETE MUST DO A FULL DATASHEET
void matrixEditor(struct Matrix* mat,const char* headers,const char* formats) //does not save screen
//spreadsheet-like matrix editor. headers of columns and rows can be given, and formats also.
{
	int rows,cols,r,c,r0,c0;
	unsigned int key;
	char buf[32];
	char prebuf2[32];
	int w1,w2;
	double d1;
	unsigned char* Scr2;
	
	rows=mat->rows;
	cols=mat->cols;
	r=1;c=1;
	r0=1;c0=1;

	while(1)
	{
		Bdisp_AllClr_VRAM();
		if(headers!=0)
		{
		for(w1=1;w1<=3;w1++)//draw horizontal headers
		{
			//strList(headers,',',c0+w1-2,buf);
			PrintMini(10+(w1-1)*38+3,1,(unsigned char *)buf,MINI_OVER);
			Bdisp_AreaReverseVRAM(10+(w1-1)*38+1,0,10+(w1)*38-1,5);
		};
		};
		for(w1=1;w1<=rows;w1++)//draw vertical headers
		{
			sprintf(buf,"%d",w1+r0-1);
			PrintMini(2,5+8*(w1-1)+2,(unsigned char *)buf,MINI_OVER);
			Bdisp_AreaReverseVRAM(1,5+8*(w1-1)+1,10,5+8*(w1)-1);
		};
		for(w1=1;w1<=7;w1++)//draw items
		{
			if(w1<=rows)
			{
			for(w2=1;w2<=3;w2++)
			{
				if(w2<=cols)
				{
				d1=matrixGet(mat,r0+w1-1,c0+w2-1);
				if(formats==0)
				{
					if(fabs(d1)<0.0001 && d1!=0.0) sprintf(buf,"%.2E",d1); else sprintf(buf,"%.4f",d1);
				}else
				{
					//strList(formats,',',c0+w2-2,prebuf2);
					sprintf(buf,prebuf2,d1);
				};
				PrintMini(10+(w2-1)*38+2+1,5+8*(w1-1)+2,(unsigned char *)buf,MINI_OVER);
				};
			};
			};
		};
		w1=r-r0+1;
		w2=c-c0+1;
		Bdisp_AreaReverseVRAM(10+(w2-1)*38+2,5+8*(w1-1)+1,10+(w2)*38-1,5+8*(w1)-1);//draw current cell
		if(c0+3-1<cols) PrintMini(123,0,(unsigned char*)"\xE6\x9B",MINI_OVER);
		if(c0>1) PrintMini(5,0,(unsigned char*)"\xE6\x9A",MINI_OVER);
label1:
		GetKey(&key);
		w2=isNumberKey(key);
		if(key==KEY_CTRL_LEFT)
		{
			if(c0>1 && c>1)
			{
				c0--;
				c--;
			}else if(c>1)
			{
				c--;
			};
		}else if(key==KEY_CTRL_RIGHT)
		{
			if(c-c0+1<3 && c<cols)
			{
				c++;
			}else if(c-c0+1==3 && c<cols)
			{
				c0++;
				c++;
			};
		}else if(key==KEY_CTRL_UP)
		{
			if(r0>1 && r>1)
			{
				r0--;
				r--;
			}else if(r>1)
			{
				r--;
			};
		}else if(key==KEY_CTRL_DOWN)
		{
			if(r-r0+1<7 && r<rows)
			{
				r++;
			}else if(r-r0+1==7 && r<rows)
			{
				r0++;
				r++;
			};
		}else if(key==KEY_CTRL_EXE || w2!=0)
		{
			if(w2!=0)
			{
				prebuf2[0]=w2;
				prebuf2[1]='\0';
			}else
			{
				d1=matrixGet(mat,r,c);
				if(fabs(d1)<0.0001 && d1!=0.0) sprintf(prebuf2,"%.5E",d1); else sprintf(prebuf2,"%.4f",d1);
			};
			buf[0]='\0';
			w1=r-r0+1;
			w2=c-c0+1;
			Scr2=GetScreen();
			PopUpWin(1);
			//TODOstrInput(4,4,buf,16,prebuf2,0);
			matrixSet(mat,r,c,atof(buf));
			SetScreen(Scr2);
			free(Scr2);
			c++;
			if(c>cols) {c=1;r++;};
			if(r>rows) {r=1;c=1;};
			if(c-c0>=3 || c0>c) c0=c-3+1;
			if(r-r0>=7 || r0>r) r0=r-7+1;
			if(c0<1) c0=1;
			if(r0<1) r0=1;
		}else if(key==KEY_CTRL_EXIT)
		{
			return;
		}else
		{
			goto label1; //do not refresh if a usable key is pressed
		};
	};
}
*/


void messageBox(const char* text,const char* title)
{
    struct Dialog* dlg=0;
    struct DialogItem* item;
    struct String* s;
    struct String* s1;
    int w1,n;
    s=stringInit(text,16);
    s1=stringInit("",64);
    n=stringListCount(s,'|');//counts text lines
    if(n>4) n=4;//maximum 4 lines
    dlg=dialogInit();
    item=dialogItemInit();
    for(w1=1;w1<=n;w1++)
    {
	stringListItem(s,'|',w1,s1);//text line
        dialogItemSetLabel(item,10,w1*10,stringGet(s1));dialogAppend(dlg,item);
    };
    dialogItemSetButton(item,40,50,45,10,"OK");dialogAppend(dlg,item);
    dialogExec(dlg,title,1);
    dialogItemFree(item);
    stringFree(s);
    stringFree(s1);
    dialogFree(dlg);
}

int questionBox(const char* question,int defaultYes)
{
    int w1=0;
    struct Dialog* dlg=0;
    struct DialogItem* item;
    dlg=dialogInit();
    item=dialogItemInit();
    dialogItemSetLabel(item,1,30,question);
    dialogAppend(dlg,item);
    if(defaultYes==1)
    {
        dialogItemSetButton(item,15,45,40,10,"Yes");dialogAppend(dlg,item);
	dialogItemSetButton(item,75,45,40,10,"No");dialogAppend(dlg,item);
    }else
    {
        dialogItemSetButton(item,15,45,40,10,"No");dialogAppend(dlg,item);
	dialogItemSetButton(item,75,45,40,10,"Yes");dialogAppend(dlg,item);
    }
    dialogExec(dlg,"Question:",1);
    w1=dlg->retCode;
    dialogItemFree(item);
    dialogFree(dlg);
    if(w1==3 && defaultYes==0) return 1;
    if(w1==2 && defaultYes==1) return 1;
    return 0;
}

int multipleChoiceBox(struct StringList* items,int curItem) 
{
	unsigned int key;
	int i0,i,n,n0;
	int w1,w2,w3;
	char* bufp;
	DISPBOX box;
	i=curItem;
	n=items->count;
	if(n>5) n0=5; else n0=n;
	i0=curItem;
	if(i0+n0>n) i0=n-n0+1;
	PopUpWin(n0);
	w3=(8-n0)/2-(!(n0%2));
	box.left=12;box.right=115;box.top=w3*8;box.bottom=(w3+n0)*8;
	while(1)
	{
		Bdisp_AreaClr_VRAM(&box);
		for(w1=1;w1<=n0;w1++)
		{
			bufp=slistGet(items,w1+i0-1);
			w2=strlen(bufp);
			locate(10-w2/2,w3+w1);
			Print((unsigned char *)bufp);
		};
		Bdisp_AreaReverseVRAM(box.left,(w3+i-i0)*8,box.right-7,(w3+i-i0+1)*8-1);//draw selection
		if(i0!=1)//up arrow
		{
			locate(19,w3+1);
			Print((unsigned char*)"\xE6\x9C");
		};
		if(i0+n0-1<n)//down arrow
		{
			locate(19,w3+n0);
			Print((unsigned char*)"\xE6\x9D");
		};
		GetKey(&key);
		if(key==KEY_CTRL_DOWN)
		{
			i++;
			if(i>n) i=n;
			if(i-i0+1>n0) i0++;
		}else if(key==KEY_CTRL_UP)
		{
			i--;
			if(i<1) i=1;
			if(i-i0<0) i0--;
		}else if(key==KEY_CTRL_EXE)
		{
			break;
		};
	};
	return i;
}



  