/*
mylib helper library for Casio 9860 sdk v0.45
Copyright (C) 2011 Toussis Manolis (koppermind at gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#ifndef MYLIB_H
#define MYLIB_H

#define PI 3.1415926535

//comment the following line to remove debug macros and asserts in code 
#define MYLIB_DEBUG
#ifdef MYLIB_DEBUG
extern char debugB[128]; //debugBuf
extern unsigned int debugkey; 
extern unsigned char* debugScr;
#define ASSERT(x) if((x)==0) {Bdisp_AllClr_VRAM();sprintf(debugB,"Check:"#x);PrintMini(1,1,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"%s",__FILE__);PrintMini(1,20,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"line: %d",__LINE__);PrintMini(1,30,(unsigned char *)debugB,MINI_OVER);GetKey(&debugkey);abort();};
#define DEBUGNUM(x) if(1) {debugScr=GetScreen();Bdisp_AllClr_VRAM();sprintf(debugB,"Debug:"#x);PrintMini(1,1,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"int:%d",x);PrintMini(1,10,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"dbl:%f",x);PrintMini(1,20,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"hex:%8X",x);PrintMini(1,30,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"%s",__FILE__);PrintMini(1,40,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"line:%d",__LINE__);PrintMini(1,50,(unsigned char *)debugB,MINI_OVER);GetKey(&debugkey);SetScreen(debugScr);free(debugScr);};
#define DEBUGSTR(x) if(1) {debugScr=GetScreen();Bdisp_AllClr_VRAM();sprintf(debugB,"Debug:"#x);PrintMini(1,1,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"%s",x);PrintMini(1,10,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"%s",__FILE__);PrintMini(1,20,(unsigned char *)debugB,MINI_OVER);sprintf(debugB,"line:%d",__LINE__);PrintMini(1,30,(unsigned char *)debugB,MINI_OVER);GetKey(&debugkey);SetScreen(debugScr);free(debugScr);};
#else
#define ASSERT(x) (1)
#define DEBUGNUM(x) (1)
#define DEBUGSTR(x) (1)
#endif //MYLIB_DEBUG

#define EVENT(object,eventtype,data) if(object->event!=0){object->eventType=eventtype;object->eventData=data;object->event(object);};

int GetMiniTextWidth(const char* text);//returns the number of pixels, text is in PrintMini
int GetOSVersion(void); //returns an int for os version
int GetRealTimeTicks(void);//returns 1/64s of second

unsigned char* GetScreen(void);//returns a buffer of current screen. Must use free(buffer) to free memory
void SetScreen(unsigned char* p);//sets the screen from buffer p


//the following line sets the size of char buf[] that is being used by stringSprintf and other string functions
//if you want you can lower the value, or raise it for manipulating bigger strings. If the code you use exceeds the boundaries 
//for that function an ASSERT() is being raised.
#define STRING_SPRINTF_SIZE 512
struct String //1 based structure
{
  char* data;
  int maxLength;
};
struct String* stringInit(const char* text,int maxsize);
char* stringGet(struct String* str);
double stringGetDouble(struct String* str);
int stringGetInteger(struct String* str);
void stringSet(struct String* str, const char* text);
void stringAppend(struct String* str, const char* text);
void stringCopy(struct String* dest,const struct String* src);
void stringSetDouble(struct String* str, const double value,const char* format);
void stringSetInteger(struct String* str,const int value,const char* format);
int stringLength(struct String* str);
void stringFree(struct String* str);
void stringSprintf(struct String* str,const char* format,...);
void stringFormatValue(struct String* str,const char* format, const char* value);
void stringFill(struct String* str,const char c, int n);//fills str with n chars c
void stringRemoveAt(struct String* str, int i,int n);//removes n chars starting from i(1based) in str
void stringMid(struct String* str,int i,int n,struct String* dest);//returns n chars starting with i(1based) of the str
int stringIdx(struct String* str,const char c,int instance);//returns the index(1based) of the instance of char c
void stringListItem(struct String* str,const char d,int i,struct String* dest);//returns the i(1based) item str with delimiter char d
int stringListCount(struct String* str,const char d);//returns the number of items in str delimited with d
void stringListItemModify(struct String* str,const char d,int i,const char* newitem,struct String* dest);// modifies the ith(1based) item delimited with d with newitem and returns the result
int isNumberKey(unsigned int key);
int isAlphaKey(unsigned int key);
#define INPUT_NUMBER 1
#define INPUT_STRING 2
void stringInput(int colx, int rowy, struct String* s, int len, int maxLen, const char* preinput, int hasBox, int type);//input box at x,y with max chars=len returns input in buf




#define DIALOG_LABEL 1
#define DIALOG_BUTTON 2
#define DIALOG_EDITBOX 3
#define DIALOG_CHECKBOX 4
#define DIALOG_RADIO 5
#define DIALOG_DROPLIST 6
#define DIALOG_LIST 7
#define DIALOG_LISTEDIT 8
#define DIALOG_MENU 9

struct DialogItem
{
  int type;//1 LABEL, 2 BUTTON, 3 EDITBOX, 4 CHECKBOX, 5 RADIO, 6 DROPLIST, 7 LIST,8 LISTEDIT,9 MENU
  int visible;
  int x,y,w,h,status;//status is used in checkbox
  struct String* text;
  struct String* format;
  struct StringList* items;
  struct StringList* formats;
  int current;
  struct StringList* values;//values is used only in LISTEDIT for now
  struct DialogItem* next;
};
struct DialogItem* dialogItemInit();
void dialogItemFree(struct DialogItem* item);
void dialogItemCopy(struct DialogItem* dest,const struct DialogItem* source,int nextlink);
void dialogItemSetLabel(struct DialogItem* itm,int x,int y,const char* text);
void dialogItemSetButton(struct DialogItem* itm,int x,int y,int w,int h,const char* text);
void dialogItemSetEditBox(struct DialogItem* itm,int x,int y,int w,int h,const char* text,const char* format);
void dialogItemSetCheckBox(struct DialogItem* itm,int x,int y,const char* text,int status);
void dialogItemSetRadio(struct DialogItem* itm,int x,int y,int current,const char* items);
void dialogItemSetDropList(struct DialogItem* itm,int x,int y,int w,int h,int current,const char* items);
void dialogItemSetList(struct DialogItem* itm,int x,int y,int w,int h,int current, const char* items);
void dialogItemSetListEdit(struct DialogItem* itm,int x,int y,int w,int h,int current,const char* labels,const char* values,const char* formats,int readOnly);
void dialogItemSetMenu(struct DialogItem* itm,const char* menu);

typedef void(*DialogCallback)(void*);//definition of a function callback must be void func(void* ptr). ptr points to Dialog*
#define EVENT_ITEM_NEXT 1
#define EVENT_ITEM_PREV 2
#define EVENT_ITEM_CHANGED 3
#define EVENT_BEFORE_REFRESH 4
#define EVENT_AFTER_REFRESH 5
#define EVENT_KEY_PRESS 6

struct Dialog//Linked List of DialogItem*
{
  struct DialogItem* first;
  int count;
  int current;
  int retCode,retSubcode;
  DialogCallback event;
  int eventType;
  int eventData;
  struct String* title;
};

void dialogPrepend(struct Dialog* dlg,const struct DialogItem* item);
struct Dialog* dialogInit();
struct DialogItem* dialogGet(struct Dialog* dlg,int i);
void dialogSet(struct Dialog* dlg, int i, struct DialogItem* item);
void dialogFree(struct Dialog* dlg);
void dialogInsertAt(struct Dialog* dlg,int i,const struct DialogItem* item);
void dialogAppend(struct Dialog* dlg,const struct DialogItem* item);
void dialogRemoveAt(struct Dialog* dlg,int i);
void dialogRemoveAll(struct Dialog* dlg);
void dialogExec(struct Dialog* dlg,const char* title,int clearScreen);


struct StringList//Linked List of char* (1 based)
{
  struct StringListNode* first;
  int count;  //number of items
  struct StringListNode* current;
};

struct StringListNode 
{
  char* data;
  int maxCount;
  struct StringListNode* next;
};

struct StringListNode* slistCreateNode(const char* text);
void slistPrepend(struct StringList* list,const char* text);
struct StringList* slistInit(int n,int itemsize);
struct StringListNode* slistFirst(struct StringList* list);
struct StringListNode* slistNext(struct StringList* list);
void slistNodeSet(struct StringListNode* item,const char* text);
char* slistGet(struct StringList* l,int i);
void slistSet(struct StringList* l,int i,const char* text);
void slistSetAll(struct StringList* l,char d,const char* items);
void slistGetAll(struct StringList* l,char d,struct String* items);
void slistCopy(struct StringList* dest,const struct StringList* src);
void slistFree(struct StringList* l);
void slistRemoveAll(struct StringList* l);
void slistInsertAt(struct StringList* l,int i,const char* text);
void slistAppend(struct StringList* l,const char* text);
void slistRemoveAt(struct StringList* l,int i);



struct Matrix //1based matrix
{
	double* data;
	int rows,cols;
};
struct Matrix* matrixInit(struct Matrix* mat,int rows,int cols);
void matrixResize(struct Matrix* mat,int rows,int cols);
void matrixSetAll(struct Matrix* mat,double** values);
double matrixGet(struct Matrix* mat,int r,int c); //returns mat[r,c]
void matrixSet(struct Matrix* mat,int r,int c,double v); //mat[r,c]=v
void matrixFree(struct Matrix* mat);
void matrixZero(struct Matrix* mat);//fills mat with 0.0
struct Matrix* matrixAppendRow(struct Matrix* mat);
struct Matrix* matrixRemoveRow(struct Matrix* mat, int r);//remove row r
struct Matrix* matrixRemoveColumn(struct Matrix* mat,int c);//remove col c
void matrixAddMatrix(struct Matrix* mat1,struct Matrix* mat2);//adds mat2 to mat1
void matrixAddNumber(struct Matrix* mat1,double v); // mat1+v
struct Matrix* matrixMulMatrix(struct Matrix* mat1,struct Matrix* mat2);//mat1 x mat2 returns result;
struct Matrix* matrixCopy(struct Matrix* mat1);//returns a new matrix with mat1 contents
struct Matrix* matrixTranspose(struct Matrix* mat1);//returns a new matrix with transpose of mat1
void matrixMulNumber(struct Matrix* mat1,double v);// mat1*v
void matrixSwapRow(struct Matrix* mat1,int r1,int r2);//swaps rows r1 and r2
void matrixAddMultiplyRow(struct Matrix* mat1,int r1,int r2,double v);//adds row r2*v to row r1
void matrixDisplay(struct Matrix* a);//for debug purposes
struct Matrix* matrixLinearSolve(struct Matrix* a,struct Matrix* b); // A X = B returns B with solution
struct Matrix* matrixConjGradient(struct Matrix* a,struct Matrix* b,double residual); //cojugate gradient method
double matrixDeterminant(struct Matrix* a);


//void matrixEditor(struct Matrix* mat,const char* headers,const char* formats); OBSOLETE TODO
int questionBox(const char* question,int defaultYes);
int multipleChoiceBox(struct StringList* items,int curItem);
void messageBox(const char* text,const char* title);

int execMenu(const char* items);//example execMenu("op1,op2,op3,op4,op5,next") returns selection

int execMenuPopup(struct String* items,int pos); //example execMenuPopup("op1-1|op1-2|op1-3",2)

#define STYLE_SOLID 0xFF
#define STYLE_DOT 0xAA
#define STYLE_DOT2 0x88
#define STYLE_DASH 0xCC
#define STYLE_DASH2 0xF0
#define STYLE_DASHDOT 0xE4
#define STYLE_DASHDOT2 0xEA
void drawCircle(int x0, int y0, int r, int color);
void drawLine(int x1, int y1, int x2, int y2, int color, int pattern);
void drawRect(int x1,int y1,int x2,int y2, int color,int pattern);
void fillRect(int x1,int y1,int x2,int y2, int color);


struct View {
	double leftX,rightX,topY,bottomY;
	double pixeldX,pixeldY;//the size of 1px in view coords
	double dx,dy;//the size of view
};
void viewPointToScreen(double x,double y,int* xi,int* yi, struct View* v);//returns xi,yi from view x,y
void viewPointToView(int xi,int yi,double* x, double* y,struct View* v);//returns x,y form screen xi,yi
void viewSet(double leftx,double topy,double rightx,double bottomy,struct View* v);//set View
void viewSetOrigin(int x0,int y0,double dx,double dy,struct View* v);//sets view with screen origin x0,y0 with view size dx,dy
void viewPoint(double x,double y,int color,struct View* v);//draws point x,y in view v
void viewLine(double x1,double y1, double x2, double y2,int color,int pattern,struct View* v);//draws line
void viewRect(double x1,double y1, double x2, double y2,int color,int pattern,struct View* v);//draws rect
void viewFillRect(double x1,double y1, double x2, double y2,int color,struct View* v);//draws rect
void viewScale(double x0,double y0,double scale,struct View* v);//scales the view 
void viewArrow(double x1,double y1,double x2, double y2, int size, int color, int pattern,struct View* v);//draws an arrow from x1,y1 to x2,y2 with arrow size pixels 
#define ALIGN_LEFT 1
#define ALIGN_CENTER 2
#define ALIGN_RIGHT 4
#define ALIGN_VERTICAL 8
#define ALIGN_VLEFT 9
#define ALIGN_VCENTER 10
#define ALIGN_VRIGHT 12
void viewPrintMini(double x,double y,const char* str,int type,int align,struct View* v);

double lineAngle(double x1,double y1,double x2,double y2); //returns angle of line

#endif
